#
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $ 
# $Revision: 1.12.5.56 $ 
# $Source: /project/arrayrep/src/srdf/unix/SRDFAgent.pm,v $ 
#

package SRDFAgent;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use POSIX qw(uname);
use File::Basename;
use ag_i18n_inc;
my $vcs_home=$VCSHOME;
my $vcs_log=$VCSLOG;
my $USERLANG=$ENV{"LANG"};
# supported microcode version
use constant SUPPORTED_MCODE_VERSION => 5568;
my ($OS, $DEVNULL) = ((uname())[0], "/dev/null");

my $VXDISK = "/usr/sbin/vxdisk";
my $powermt = '/sbin/powermt';
my $blockdev = '/sbin/blockdev';
my $native_multipathing = '/sbin/multipath';

my ($symrdf, $symdg, $sympd, $symcfg, $symcg, $symdev, $symcli);

if ($OS eq "Windows NT") {
    $DEVNULL = "nul";
}

@EXPORT = qw(
);
$VERSION = '1.00';

require 5.000;

%default = (
        resname => '',
        name    => '',
        home    => '',
        mode    => '',
        state   => '',
        role    => '',
        devs    => 0,
        invalid => 0,
        devfotime => 0,
        iscompositegroup => 0,
        dcv_mode => '',
        consistency => '',
        numofRDFGs => 0,
        onlinetimeout => 0,
        swaproles => 0,
        resource_type => 'SRDF',
        ARRAYDEVS => {},
        start_time => '',
        RDFG => {},
        LSymmetrix => {},
        RSymmetrix => {},
        type => '',
        concurrent => -1,
        preferredrdfg => '',
        rdfgopt => '',
        is_vmware => 0,
        grp_type => '',
        dev_state => '',
		pairstate => '' 
);

sub new {
    my($proto, $resname, $name, $symhome, $mode, $devfotime, $iscompositegroup, $swaproles) = @_;
    my $class = ref($proto) || $proto;
    my $self = { %default };
    my @RDFG;
    my $numofRDFGs;
    my (%arraydevs)  = ();
    my $func = "SRDFAgent::new";

    $self->{'resname'} = $resname;
    $self->{'name'} = $name;
    $self->{'home'} = $symhome;
    $self->{'mode'} = $mode;
    $self->{'iscompositegroup'} = $iscompositegroup;
    $self->{'swaproles'} = $swaproles;
    $self->{'start_time'} = $^T;

    if ($self->{'mode'} eq ''){
        $self->{'mode'} = "MODE_NOT_SPEC";
    }
    if ($self->{'iscompositegroup'} == 0){
        $self->{'type'} = " -g ";
    } else {
        $self->{'type'} = " -cg ";
    }

    my $osType = $^O;
    if ( $osType =~ /^MSWin/i)
    {
        $self->{'home'} = "\"" . Win32::GetShortPathName($symhome) . "\"";
        $vcs_home = "\"" . Win32::GetShortPathName($VCSHOME) . "\"";
        $symrdf = "$self->{'home'}\\symrdf";
        $symdg = "$self->{'home'}\\symdg";
        $sympd = "$self->{'home'}\\sympd";
        $symcfg = "$self->{'home'}\\symcfg";
        $symcg = "$self->{'home'}\\symcg";
        $symdev = "$self->{'home'}\\symdev";
        $symcli = "$self->{'home'}\\symcli";
    }
    else
    {
        $self->{'home'} = $symhome;
        $symrdf = "$self->{'home'}/bin/symrdf";
        $symdg = "$self->{'home'}/bin/symdg";
        $sympd = "$self->{'home'}/bin/sympd";
        $symcfg = "$self->{'home'}/bin/symcfg";
        $symcg = "$self->{'home'}/bin/symcg";
        $symdev = "$self->{'home'}/bin/symdev";
        $symcli = "$self->{'home'}/bin/symcli";
    }

    $self->{'consistency'} = "Disabled";
    $self->{'devfotime'} = $devfotime;
    $self->{ARRAYDEVS} = \%arraydevs;

    $self->{'concurrent'} = -1;
    $self->{'rdfgopt'} = "";

    bless($self, $class);
    return $self;
}

sub DESTROY {}

sub _exit
{
    my ($self, $stat) = @_;
    $self->unlock_res();
    exit($stat);
}

sub Rep_Tech_RPO {
    my $ResName = shift;
    my ($groupname, $iscompositegroup, $symhome, $time, $ret, $time_behind);

    ($ret, $symhome) = VCSAG_GET_ATTR_VALUE ("SymHome", -1, 1, @ARGV);
    exit 0 if ($ret != $VCSAG_SUCCESS);
    ($ret, $groupname) = VCSAG_GET_ATTR_VALUE ("GrpName", -1, 1, @ARGV);
    exit 0 if ($ret != $VCSAG_SUCCESS);
    ($ret, $iscompositegroup) = VCSAG_GET_ATTR_VALUE ("IsCompositeGroup", -1, 1, @ARGV);
    exit 0 if ($ret != $VCSAG_SUCCESS);

    if ($iscompositegroup){
        $time_behind = `$symhome/bin/symcg show $groupname | grep 'Time that R2 is behind R1'`;
    } else {
    	$time_behind = `$symhome/bin/symdg show $groupname | grep 'Time that R2 is behind R1'`;
    }  	
    $time_behind =~ /\s+Time\ that\ (.*\w+)\s+:\s+(.*:.*:.*)\s+$/;
    $time = "Rep_Tech_RPO: "."$1"." by"." $2"." [hh:mm:ss] time as reported by Symmetrix.";
    return $time;
}

#Populates all the path information in the the devices array for a $dev. On linux powerpath commands are installed at /sbin/. 
#On other platform path may differ.
sub get_all_paths {
    my $func = 'get_all_paths';
    VCSAG_LOGDBG_MSG(1, "$func() Entered");
    my $dev = shift;
    my $devices = shift;
    my $powermt_cmd = "$powermt display dev=$dev";
    my @powermt_output;
    run_cmd ($powermt_cmd, \@powermt_output, 1, $func);
    foreach my $line (@powermt_output) {
        next if ($line =~ m /^$|^Pseudo|^Symmetrix|^Logical|^state|^=|^#|^-/);
        my @dev_array = split (/\s+/, $line);
        push (@$devices, $dev_array[3]);
    }
    VCSAG_LOGDBG_MSG(1, "$func() Exited");
}

# First check for type of multipathing.
# After multipathing is found get all the devices associated.
sub reload_devicemap {
    my $self = shift;
    my $func = "reload_devicemap";
    return if ($^O !~ /linux/i);
    VCSAG_LOGDBG_MSG(1, "$func() Entered");
    my $native_multipath_cmd = "$native_multipathing -ll";
    my $powermt_version = "$powermt version 1> $DEVNULL";
    my $multipath_output;
    my $updated=0;
    VCSAG_LOGDBG_MSG(1, "Checking for native multipathing.");
    run_cmd($native_multipath_cmd, \$multipath_output, 1, $func);
    if ($multipath_output && $? == 0) {
        VCSAG_LOGDBG_MSG(1, "Native multipathing found.");
        my @pdevs = $self->get_pdev_name();
        foreach my $dev (@pdevs){
            if (grep ($dev, $multipath_output)){
                my $cmd = "echo 1 > /sys/block/$dev/device/rescan";
                run_cmd ($cmd, 0, 1, $func);
                $updated = 1;
            }
        }
    } else {
        VCSAG_LOGDBG_MSG (1, "Native multipathing not found.");
    }
    if ($updated) {
        run_cmd ("multipath -r", 0, 1, $func);
    }
    VCSAG_LOGDBG_MSG(1, "Checking for powerpath.");
    run_cmd ($powermt_version, 0, 1, $func);
    if ($? == 0) {
        VCSAG_LOG_MSG ("I", "Powerpath found", 1112);
        my @pdevs = $self->get_pdev_name();
        foreach my $dev (@pdevs) {
    	    my @devices;
            get_all_paths($dev, \@devices);
            s/^/\/dev\// for @devices;
            push (@devices, "/dev/$dev");
            foreach my $dev_path (@devices) {
                my $cmd = "$blockdev --rereadpt $dev_path";
                run_cmd ($cmd, 0, 1, $func);
            }
        }
    } else {
        VCSAG_LOGDBG_MSG (1, "Powerpath not found.");
    }
    VCSAG_LOGDBG_MSG(1, "$func() Exited");
}

sub get_pdev_name {
    my $self = shift;
    my $cmd;
    if ($self->{'iscompositegroup'} == 0) {
        $cmd = "$symdg show $self->{'name'} | grep '\/dev.*RDF'";
    } else { 
        $cmd = "$symcg show $self->{'name'} | grep '\/dev.*RDF'";
    }
    my (@dg_show_output, @pdevs);
    my $func = "get_pdev_name";
    VCSAG_LOGDBG_MSG(1, "$func() Entered");
    run_cmd($cmd, \@dg_show_output, 1, $func);
    foreach my $line (@dg_show_output) {
        my @dev_array = split (/\s+/, $line);
        my @device_path = split(/\//,$dev_array[2]);
        push (@pdevs, $device_path[-1]);
    }
    VCSAG_LOGDBG_MSG(1, "Following pdevs are associated with the group [@pdevs].");
    VCSAG_LOGDBG_MSG(1, "$func() Exited");
    return @pdevs;
}

sub get_state {
    my $self = shift;
    my $ret;
    my $indexRDFG = 0;
    my $indexLSymmID = 0;
    my $indexRSymmID = 0;
    my $key;
    my $devidx = 0;
    my $output;
    my $func = "get_state";
    my $cmd = "$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} query";
    run_cmd($cmd, \$output, 1, $func);
    my @out;

    if(!$output) {
        $ret = "$?";
        my $cmd_txt=$cmd;
        $cmd_txt =~ s/\"//g;
        VCSAG_LOG_MSG("E", "Command $cmd_txt failed with error code:$ret\n", 1008, "$cmd_txt", "$ret");
        $self->_exit(0);
    }

    my @query = split('\n', $output);
    $self->{'state'} = "";
    foreach my $q (@query) {
        if ($q =~ /Group Number/) {
            @out = split('\s+', $q);
            $self->{RDFG}{$indexRDFG} = $out[5];
            $indexRDFG++;
            $devidx = 0;
            next;
        }
        if ($q =~ /^Remote Symmetrix ID/) {
            @out = split('\s+', $q);
            $self->{RSymmetrix}{$indexRSymmID} = $out[4];
            $indexRSymmID++;
            next;
        }
#
# a local symmetrix id can appear in these forms:
# DG's Symmetrix ID   : 000190300835    (Microcode  Version: 5773)
# OR
# Symmetrix ID        : 000190300835    (Microcode  Version: 5773)
#
        if ($q =~ /^Symmetrix ID/ or $q =~ /^DG's Symmetrix ID/) {
            my @out2 = split(':', $q);
            @out = split('\s+', $out2[1]);
            $self->{LSymmetrix}{$indexLSymmID} = $out[1];
            $indexLSymmID++;
            next;
        }
        if ($q =~ /Source/) {
            my @roleinfo = split('\s+', $q);
            if ($roleinfo[1] eq "Source") {
                $self->{'role'} = "r1";
            } else {
                $self->{'role'} = "r2";
            }
            next;
        }
        if (!($q =~/[SAC]\./)) {
            next;
        }

        my @stateinfo = split('\s+', $q);

        if ($stateinfo[10] =~ /S/) {
            $self->{'dcv_mode'} = "Synchronous";
        } elsif ($stateinfo[10] =~ /A/) {
            $self->{'dcv_mode'} = "Asynchronous"; 
        }

        if ($self->{'iscompositegroup'} == 1) {
            if ($stateinfo[11] =~ /X/) {
                $self->{'consistency'} = "Enabled";
            } elsif ($stateinfo[11] =~ /./) {
                $self->{'consistency'} = "Disabled";
            } 
            $key = "Local" . $self->{RDFG}{$indexRDFG - 1} .  $self->{LSymmetrix}{$indexLSymmID - 1};
            $self->{'ARRAYDEVS'}{$key}{$devidx} = $stateinfo[1];
            $key = "Remote" . $self->{RDFG}{$indexRDFG - 1} .  $self->{RSymmetrix}{$indexRSymmID - 1};
            $self->{'ARRAYDEVS'}{$key}{$devidx} = $stateinfo[6];
            $devidx++;
        }

        if ($stateinfo[2] eq "RW") {
            if ($self->{'state'} eq "WD") {
                $self->{'state'} = "Invalid";
                return;
            } else {
                $self->{'state'} = "RW";
            }
        }
        if ($stateinfo[2] eq "WD") {
            if ($self->{'state'} eq "RW") {
                $self->{'state'} = "Invalid";
                return;
            } else {
                $self->{'state'} = "WD";
            }
        }

        ++$self->{'devs'};
        $self->{'invalid'} = $self->{'invalid'} + $stateinfo[3];
    }
    if (!$self->{'state'}) {
        $self->{'state'} = "Invalid";
    }
    if ($self->{'iscompositegroup'} == 1) {
        $self->{'numofRDFGs'} = $indexRDFG;
    }
}

sub get_mode {
    my $self = shift;
    my $retVal = 0; # 0 stands for Synchronous
    my $dcv_mode = $self->{'dcv_mode'};
    my $func = "get_mode";

	# Check the specified mode with the discovered mode. The two should match. 
	# If not, then give an error and exit.
	#
	# Partitioned state is a very special case. 
	#
	# In partitioned state, for asynchronous mode of replication, the mode on
	# the R2 side changes to Synchronous (actually it should drop to adaptive
	# copy, but it drops to the default of "S" i.e. synchronous.)
	#
	# To consider the above case, we do not signal an error if -
	#
	# - the devices are in a partitioned state.
	# - The mode specified in the configuration is 'Asynchronous'
	# - The discovered mode is 'Synchronous'
	#
	# However, in a partitioned state, we will signal an error if -
	#
	# 1. The configured state is 'Synchronous' and the discovered state 
	#    is 'Asynchronous'
	#
	# +-------------------------------------------------------------+
	# |   Configured State  |  Discovered State | Result            |
	# |---------------------+-------------------+-------------------|
	# |                     |                   |                   |
	# |   Synchronous       | Synchronous       | Valid             |
	# |---------------------+-------------------+-------------------|
	# |   Synchronous       | Asynchronous      | Error             |
	# |---------------------+-------------------+-------------------|
	# |   Asynchronous      | Synchronous       | Valid if          |
	# |                     |                   | partitioned state |
	# |                     |                   |                   |
	# |                     |                   | Error if not      |
	# |                     |                   | partitioned state |
	# |                     |                   |                   |
	# |---------------------+-------------------+-------------------|
	# |   Asynchronous      | Asynchronous      | Valid             |
	# +-------------------------------------------------------------+
	#
	if ($self->{'mode'} ne "MODE_NOT_SPEC") {

        if (($self->{'mode'} eq $dcv_mode) && ($self->{'mode'} eq "Asynchronous")) {
            $retVal = 1;
        }

        if (($self->{'mode'} ne $dcv_mode) && ($self->{'mode'} eq "Synchronous")) {
# Error - Configured is Sync and discovered is Async.
            VCSAG_LOG_MSG("E", "Inconsistent mode of replication detected. Configured is $self->{'mode'}, detected is $dcv_mode", 24, "$self->{'mode'}", "$dcv_mode");
            $self->_exit(0);
        }
        if (($self->{'mode'} ne $dcv_mode) && ($self->{'mode'} eq "Asynchronous")) {
# Configured is Asynch and discovered is sync.
            my $partitioned = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -partitioned verify > $DEVNULL 2>&1", 0, 1, $func);
            if ($partitioned != 0) {
# Error if not partitioned.
                VCSAG_LOG_MSG("E", "Inconsistent mode of replication detected. Devices in group $self->{'name'} are not in partitioned state, mode of replication configured is $self->{'mode'}, detected is $dcv_mode", 25, "$self->{'name'}", "$self->{'mode'}", "$dcv_mode");
                $self->_exit(0);
            }
            VCSAG_LOG_MSG("N", "Devices in group $self->{'grpname'} are in partitioned state, mode of replication configured is $self->{'mode'}, detected is $dcv_mode", 26, "$self->{'name'}", "$self->{'mode'}", "$dcv_mode");
            $retVal = 1;
        }
    } else {
        if ($dcv_mode eq "Asynchronous") {
            $retVal = 1;
        }
    }
    return $retVal;
}

sub get_consistency {
    my $self = shift;
    my $ret = 0;
    my $func = "get_consistency";
    my @query;
    if ($self->{'iscompositegroup'} == 0){
        run_cmd("$symdg show $self->{'name'}", \@query, 1, $func);
        foreach my $q (@query) {
            if (($q =~ /Session Consistency State/) && ($q =~ /Enabled/)) {
                return 1;
            }
        }
    }
    return 0;
}

sub failover_required {
    my ($self) = shift;
    my $func = "failover_required";
    if ($self->{'state'} eq "Invalid") {
        return 2;
    }

    if (($self->{'state'} eq "RW") && ($self->{'role'} eq "r2")) {
        my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -split verify >$DEVNULL 2>&1", 0, 1, $func);
        if ($ret == 0) {
            return 3;
        }
    }
    if ($self->{'state'} eq "WD") {
        my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -suspended verify >$DEVNULL 2>&1", 0, 1, $func);
        if ($ret == 0) {
            return 4;
        }
        return 1;
    }
    return 0;
}

sub failover_time {
    my ($self) = @_;
    my $cur_time = time();
    my $remaining_time = ($self->{'onlinetimeout'} - ($cur_time - $self->{'start_time'}));
    my $failover_time = $self->{'devs'} * $self->{'devfotime'};
    my $count = int(($remaining_time - ($failover_time + 10)) / 5);
    if ($count) {
        $count = $count - 1;
    }
    return $count;
}

sub get_count {
    my ($self) = @_;
    my $cur_time = time();
    my $func = "get_count";
	#
	# Get the resource-specific OnlineTimeout if set, else get the SRDF
	# type specific value
	#
    if ($self->{'onlinetimeout'} == 0) {
        my @onlinetimeout;
        if (run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout", 
                    \@onlinetimeout, 1, $func)) {
            run_cmd("$ag_i18n_inc::HATYPE -value SRDF OnlineTimeout", 
                    \@onlinetimeout, $func);
        }
        $self->{'onlinetimeout'} = $onlinetimeout[0];
    }
    my $remaining_time = ($self->{'onlinetimeout'} - ($cur_time - $self->{'start_time'}));
    my $count = int($remaining_time / 5) - 1;
    if ($count <= 0) {
        VCSAG_LOG_MSG("W", "not enough time to issue the symcli commands; administrative intervention or automatic restart required", 60);
        $self->_exit(0);
    }
    return $count;
}

sub wait_for_sync {
    my ($self) = @_;
    my $count = $self->get_count();
    my $halfCount;
    my $func = "wait_for_sync";
    if ($self->{'role'} eq "r1") {
        my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -updateinprog verify >$DEVNULL 2>&1", 0, 1, $func);
        $ret = $ret / 256;
        if ($ret != 36) {
            VCSAG_LOG_MSG("N", "failback of group $self->{'name'} will wait until all updates in progress complete or entry point times out", 75, "$self->{'name'}");
            my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $count -i 5 -updated verify >$DEVNULL 2>&1", 0, 1, $func);
            if ($ret) {
                return 1;
            }
        }
    } else {
        my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -syncinprog verify >$DEVNULL 2>&1", 0, 1, $func);
        $ret = $ret / 256;
        if ($ret != 28) {
            VCSAG_LOG_MSG("N", "failover of group $self->{'name'} will wait until all synchronization in progress complete or entry point times out", 62, "$self->{'name'}");
            if ($self->get_mode()){
                $halfCount = int($count / 2);
                $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -consistent verify >$DEVNULL 2>&1", 0, 1, $func);
                if ($ret) {
                    $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -synchronized verify >$DEVNULL 2>&1", 0, 1, $func);
                }
            } else {
                $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $count -i 5 -synchronized verify >$DEVNULL 2>&1", 0, 1, $func);
            }

            if ($ret) {
                return 1;
            }
        }
    }
    return 0;
}

# Delete the vmdisks corresponding to the disks in the SRDF group.
sub delete_vm_disks
{
    my ($self) = @_;
    my $func = "delete_vm_disks";
    if ($OS ne "AIX") {
        return;
    }
# We cannot get the symdev - OS Handle mapping from remote symapi server.
# Hence return from here if SYMCLI_CONNECT is set.
    if ($ENV{SYMCLI_CONNECT}) {
        VCSAG_LOG_MSG("I", "Skipping processing of vmdisks on AIX since SYMCLI_CONNECT is set.", 50);
        return;
    }

    VCSAG_LOGDBG_MSG(1, "$func() Entered");
    my @local_symms = ();
    my %sym_pdevs = ();
    for my $symm (keys %{$self->{LSymmetrix}}) {
        my $sid = $self->{LSymmetrix}{$symm};
        push(@local_symms, $sid);

# Collect 'sympd list' o/p for each array.
        my @sympd_lines;
        run_cmd("$self->{'home'}/bin/sympd list -sid $sid |awk '/^\\/dev\\/[a-zA-Z0-9_]* /{print \$2,\$1}'",
                 \@sympd_lines, 1, $func); #Eliminating the multipathed devices. Capturing only the devices in /dev/ directory to avoid dupliactes entries of the same device
        s/rhdisk/hdisk/g foreach (@sympd_lines); # Replace rhdisk with hdisk since vxdisk list contains hdisk only.
            $sym_pdevs{$sid} = [ @sympd_lines ];
    }
# Find symm devices in each Symmetrix participating in the RDF group.
    my %grpdevs_hash = ();
    if(! $self->_get_grp_devs($self->{'type'}, \%grpdevs_hash, @local_symms)) {
        VCSAG_LOG_MSG("E", "delete_vm_disks: Could not find devices specific to RDF Group.", 86);
        return;
    }
# Find the pdev names of the devices participating in the RDF group.
    my @all_paths = ();
    for my $sid (keys %grpdevs_hash) {
        my $dev_list_pat = join('|', @{$grpdevs_hash{$sid}});
        my @pdevs = grep (/$dev_list_pat/, @{$sym_pdevs{$sid}});
        s/\S+\s+.*(hdisk.*)/^$1 / foreach (@pdevs); # Extract the pdev name from the path and create a pattern for matching in the vxdisk list o/p.;
        my @pdevs_raw = grep (/hdisk/, @pdevs);
        push(@all_paths, @pdevs_raw);
    }

    my @vxdisks;
    my ($awk_buff, $itr) = (399, 0);
    my $awk_regex = $all_paths[0] if ($#all_paths != -1);
    $itr++;

    # awk fails if more then 399 bytes are present in the regex on AIX. To avoid the failure processing in batches
    while ($itr < ($#all_paths + 1)) {
        my $disk_len = length ($all_paths[$itr]);
        if (length($awk_regex) + 2 + $disk_len <= $awk_buff){
            $awk_regex .= " |$all_paths[$itr]";
            $itr++;
            if ($itr >= ($#all_paths +1)){
                my @temp_arr;
                run_cmd("$VXDISK path |awk '/$awk_regex/ {print \$2}' |sort -u", \@temp_arr, 1, $func);
                push (@vxdisks, @temp_arr);
            }
        } else {
            my @temp_arr;
            run_cmd("$VXDISK path |awk '/$awk_regex/ {print \$2}' |sort -u", \@temp_arr, 1, $func);
            push (@vxdisks, @temp_arr);
            $awk_regex = $all_paths[$itr];
            $itr++;
        }
    }
    my ($buf_len, $i, $op) = (256, 0);
    my $cmd = "$VXDISK rm ";
    my $cmd_len = length($cmd);
    my $num_disks = $#vxdisks + 1;
    while ($i < $num_disks) {
        my $dn_l = length($vxdisks[$i]) + 1;
        if (($cmd_len + $dn_l) <= $buf_len) {
            $cmd .= "$vxdisks[$i] ";
            $cmd_len += $dn_l;
            $i++;
            if ($i >= $num_disks) {
                my $ret = run_cmd($cmd, \$op, 1, $func);
                if ($ret) {
                    VCSAG_LOG_MSG("W", "vxdisk removal command [$cmd] failed with error code: [$ret], o/p: [$op]", 88, "$cmd", "$ret", "$op");
                }
                last;
            }
        } else {
            my $ret = run_cmd($cmd, \$op, 1, $func);
            if ($ret) {
                VCSAG_LOG_MSG("W", "vxdisk removal command [$cmd] failed with error code: [$ret], o/p: [$op]", 88, "$cmd", "$ret", "$op");
            }
            $cmd = "vxdisk rm";
            $cmd_len = length($cmd);
        }
    }
    VCSAG_LOGDBG_MSG(1, "$func() Exiting.");
}

# This function deletes the vmdisks corresponding to the luns in the device
# group and then rescanning the vxvm subsystem to recreate the lost vmdsks.
# This is done only on AIX since on AIX, vxvm caches the "readonly"ness of
# the devices even after failing over.
sub vxvm_scan
{
    my ($self) = @_;
    my $func = "vxvm_scan";
    if ($OS ne "AIX") {
        return;
    }
    $self->delete_vm_disks();
    run_cmd("$VXDISK scandisks", 0, 1, $func);

    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    my ($para) = split('\n', `$ag_i18n_inc::HAGRP -value $grp Parallel`);
    return 0 if ($para != 1);

# Find all nodes in the SystemList and issue vxvm_rescan on them
    my (@locsysarr);
    run_cmd("$ag_i18n_inc::HASYS -nodeid", \@locsysarr, 1, $func);
    my @tokens = split(/\s+/, $locsysarr[0]);
    my ($thisnode) = $tokens[5];

    my ($grp, $syslist_str);
    run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} Group", \$grp, 1, $func);
    run_cmd("$ag_i18n_inc::HAGRP -value $grp SystemList", \$syslist_str, 1, $func);
    my @syslist = split(/\s+/, $syslist_str);
    foreach my $sys (@syslist) {
        next if ($sys =~ /^\d+$/);
        next if ($sys eq $thisnode);
        run_cmd("$ag_i18n_inc::HARES -action $self->{'resname'} vxvm_scan -sys $sys", 0, 1, $func);
    }
}

sub check_invalid_r21
{
    my ($self, $op) = @_;

    my $ret = 0;
    if ($op =~ /The other pair of the R\d\d is not in a valid RDF state for this operation./) {
        self->validate_remote_r21();
# TODO: Check return status.
        $ret = 1;
    }
    return $ret;
}

sub run_cmd_chk_err
{
    my ($self, $cmd, $func) = @_;

    my $ret = 0;
    my $op;
    $ret = run_cmd($cmd, \$op, 1, $func);
    if($ret) {
        if ($self->check_invalid_r21($op)) {
# Reissue the failover command again.
            $ret = system($cmd);
            if ($ret) {
                $ret = $ret / 256;
            }
        } else {
            $ret = $? / 256;
        }
    }
    return $ret;
}

sub takeover {
    my ($self, $autotakeover) = @_;
    my $count;
    my $failovercount;
    my $func = "takeover";

	#
	# Get the resource-specific OnlineTimeout if set, else get the SRDF
	# type specific value
	#
    if ($self->{'onlinetimeout'} == 0) {
        my @onlinetimeout;
        if (run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout", 
                    \@onlinetimeout, 1, $func)) {
            run_cmd("$ag_i18n_inc::HATYPE -value SRDF OnlineTimeout", 
                    \@onlinetimeout, $func);
        }
        $self->{'onlinetimeout'} = $onlinetimeout[0];
    }

    if ($OS ne "Windows NT") {
        my $ret = run_cmd("/usr/bin/vmware -v", 0, 1, $func);
# If command succeeds, we will assume that it is vmware system
        if ($ret == 0) {
            $self->{'is_vmware'} = 1;
        }
    }

    my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -partitioned verify > $DEVNULL 2>&1", 0, 1, $func);
    if ($ret == 0) {
        if ($autotakeover) {
# Synchronous mode and invalid tracks found.
            if (!($self->get_mode()) && ($self->{'invalid'})) {
                VCSAG_LOG_MSG("E", "devices in group $self->{'name'} are in Partitioned state and invalid tracks exist; takeover cannot occur and administrative intervention to restore consistent data is required", 11, "$self->{'name'}");
                $self->_exit(0);
            } else {
# Asynchronous mode and invalid tracks found.
# OR
# Synchronous mode and no invalid tracks found.
                $failovercount = $self->failover_time();
                if ($failovercount <= 0) {
                    VCSAG_LOG_MSG("W", "not enough time to issue failover command; administrative intervention or automatic restart required", 27);
                    $self->_exit(0);
                }

                VCSAG_LOG_MSG("N", "devices in group $self->{'name'} are in Partitioned state, and autotakeover is enabled; failover of devices will proceed", 41, "$self->{'name'}");
                my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -c $failovercount -i 5 failover 2>&1",
                        $func);
                if($ret) {
                    VCSAG_LOG_MSG("E", "failover of Partitioned devices in group $self->{'name'} failed; administrative intervention is required", 12, "$self->{'name'}");
                    $self->_exit(0);
                }
                VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
                $self->vxvm_scan();
                $self->create_lockfile();
                if ($self->{'is_vmware'}) {
                    $self->vmware_online();
                } else { 
                    $self->cvm_import();
                }
                $self->reload_devicemap();
                $self->_exit(0);
            }
        } else {
            VCSAG_LOG_MSG("E", "devices in group $self->{'name'} are in Partitioned state but autotakeover is not enabled; administrative intervention is required", 13, "$self->{'name'}");
            $self->_exit(0);
        }
    }
}

sub failback {
    my ($self) = @_;
    my $count;
    my $halfCount;
    my $failbackcount;
    my $func = "failback";

	#
	# Get the resource-specific OnlineTimeout if set, else get the SRDF
	# type specific value
	#
    if ($self->{'onlinetimeout'} == 0) {
        my @onlinetimeout;
        if (run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout", 
                    \@onlinetimeout, 1, $func)) {
            run_cmd("$ag_i18n_inc::HATYPE -value SRDF OnlineTimeout", 
                    \@onlinetimeout, $func);
        }
        $self->{'onlinetimeout'} = $onlinetimeout[0];
    }

    if ($OS ne "Windows NT") {
        my $ret = run_cmd("/usr/bin/vmware -v", 0, 1, $func);
# If command succeeds, we will assume that it is vmware system
        if ($ret == 0) {
            $self->{'is_vmware'} = 1;
        }
    }

    $count = $self->get_count();
    $halfCount = int($count / 2);
    my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -failedover verify > $DEVNULL 2>&1", 0, 1, $func);
    if ($ret) {
        $count = $self->get_count();
        $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -updated verify > $DEVNULL 2>&1", 0, 1, $func);
        if ($ret) {
            VCSAG_LOG_MSG("E", "all devices in group $self->{'name'} are not in Failed Over or Updated state in preparation for failback; administrative intervention is required", 14, "$self->{'name'}");
            $self->_exit(0);
        }
    }
    $failbackcount = $self->failover_time();
    if ($failbackcount <= 0) {
        VCSAG_LOG_MSG("W", "not enough time to issue failback command; administrative intervention or automatic restart required", 28);
        $self->_exit(0);
    }

    VCSAG_LOG_MSG("N", "issuing failback of group $self->{'name'}", 63, "$self->{'name'}");
    $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $failbackcount failback", 0, 1, $func);

    if ($ret) {
        $ret = $ret / 256;
        VCSAG_LOG_MSG("E", "failback of group $self->{'name'} failed; error code is $ret", 16, "$self->{'name'}", "$ret");
        $self->_exit(0);
    }

    $self->get_state();
    if ($self->{'state'} eq "RW") {
        VCSAG_LOG_MSG("N", "failback of devices in group $self->{'name'} succeeded", 22, "$self->{'name'}");
        $self->vxvm_scan();
        $self->create_lockfile();
        if ($self->{'is_vmware'}) {
            $self->vmware_online();
        } else { 
            $self->cvm_import();
        }
        $self->reload_devicemap();
        $self->_exit(0);
    } else {
        VCSAG_LOG_MSG("E", "failback of group $self->{'name'} failed", 16, "$self->{'name'}");
        $self->_exit(0);
    }
}

sub validate_remote_r21 {
    my ($self) = @_;
    my $func = "validate_remote_r21";

    my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} rw_enable r1 -hop2 -noprompt", 0, 1, $func);
    if ($ret) {
        VCSAG_LOG_MSG("E", "Cannot failover the SRDF replication. The SRDF agent could not change the status of the other pair of the R21 device in device group [$self->{'name'}]. Administrative intervention required.", 82, "$self->{'name'}");
        $self->_exit(0);
    }
}

sub failover {
    my ($self) = @_;
    my $count;
    my $halfCount;
    my $failovercount;

    my $ret = 0;
    my $disabled_consistency;
    my $index;
    my $fileName;
    my $func = "failover";

	#
	# Get the resource-specific OnlineTimeout if set, else get the SRDF
	# type specific value
	#
    if ($self->{'onlinetimeout'} == 0) {
        my @onlinetimeout;
        if (run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout", 
                    \@onlinetimeout, 1, $func)) {
            run_cmd("$ag_i18n_inc::HATYPE -value SRDF OnlineTimeout", 
                    \@onlinetimeout, $func);
        }
        $self->{'onlinetimeout'} = $onlinetimeout[0];
    }

    if ($OS ne "Windows NT") {
        my $ret = run_cmd("/usr/bin/vmware -v", 0, 1, $func);
# If command succeeds, we will assume that it is vmware system
        if ($ret == 0) {
            $self->{'is_vmware'} = 1;
        }
    }

    $count = $self->get_count();
    if ($self->get_mode()) {
        $halfCount = int($count / 2);
        $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -consistent verify > $DEVNULL 2>&1", 0, 1, $func);
        if ($ret) {
            $halfCount = int($count / 2);
            $count = $self->get_count();
            $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -synchronized verify >$DEVNULL 2>&1", 0, 1, $func);
        }
    } else {
        $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $count -i 5 -synchronized verify >$DEVNULL 2>&1", 0, 1, $func);
    }
    if ($ret) {
        VCSAG_LOG_MSG("W", "all devices in group $self->{'name'} are not in synchronized or consistent state in preparation for failover; administrative intervention is required", 17, "$self->{'name'}");
        $self->_exit(0);
    }

    $failovercount = $self->failover_time();
    if ($failovercount <= 0) {
        VCSAG_LOG_MSG("W", "not enough time to issue failover command; administrative intervention or automatic restart required", 27);
        $self->_exit(0);
    }

    if ($self->{'iscompositegroup'} == 1) {
        $count = $self->get_count();

        if ($self->{'consistency'} eq "Enabled") {
            VCSAG_LOG_MSG("N", "disabling consistency for group $self->{'name'}", 76, "$self->{'name'}");
            $ret = run_cmd("$symcg $self->{'type'} $self->{'name'} -i 5 -c $count -noprompt -force disable", 0, 1, $func);
            if ($ret) {
                VCSAG_LOG_MSG("E", "Disabling the consistency for the group $self->{'name'} failed; error code is $ret", 77, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }
        } 
		

        VCSAG_LOG_MSG("N", "issuing failover of group $self->{'name'}", 18, "$self->{'name'}");
        my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $failovercount failover 2>&1", $func);
        if($ret) {
            $ret = $ret / 256;
            VCSAG_LOG_MSG("E", "Failover of the group $self->{'name'} failed; error code is $ret", 78, "$self->{'name'}", "$ret");
            $self->_exit(0);
        }

        VCSAG_LOG_MSG("N", "suspending the links for group $self->{'name'}", 61, "$self->{'name'}");
        my @out;
        $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -i 5 -c $count -noprompt suspend 2>&1", \@out, 1, $func);
        $ret = $ret>>8;
		#
		# hack to overcome bug in SE7.0.X
		# The bug is in the return code of "symrdf ...suspend". It returns a 1
		# if already in the desired state insted of CLI_C_ALREADY_IN_STATE(18)
		#
        {
            $ret = $self->__hack_se7_0_X_bug1($ret, @out);
        }

        if($ret && $ret != 18) {
            VCSAG_LOG_MSG("E", "Suspension of the links for the group $self->{'name'} failed; error code is $ret", 55, "$self->{'name'}", "$ret");
            $self->_exit(0);
        }

        if (($self->exists_swapfile()) && ($self->{'swaproles'} == 1)) {
            VCSAG_LOG_MSG("N", "swapping the roles of the devices for group $self->{'name'}", 64, "$self->{'name'}");
            for ($index=0; $index<$self->{'numofRDFGs'}; $index++) {
                $ret = $self->create_device_files(\$fileName, $self->{LSymmetrix}{$index}, $self->{RSymmetrix}{$index}, $self->{RDFG}{$index});
                if ($ret) {
                    VCSAG_LOG_MSG("E", "Error creating the device file for group $self->{'name'} ", 65, "$self->{'name'}");
                    $self->_exit(0); 
                }
                sleep 3;
                $count = $self->get_count();
                my $ret = $self->run_cmd_chk_err("$symrdf -f \"$fileName\" -rdfg $self->{RDFG}{$index} -sid $self->{LSymmetrix}{$index} swap -noprompt -i 5 -c $count", $func);
                if($ret) {
                    VCSAG_LOG_MSG("E", "Swapping the roles of the devices for the group $self->{'name'} failed; error code is $ret", 66, "$self->{'name'}", "$ret");
                    $ret = $self->remove_device_files($fileName);
                    if ($ret) {
                        VCSAG_LOG_MSG("E", "Error removing the device file for group $self->{'name'} ", 67, "$self->{'name'}");
                    }
                    $self->_exit(0);
                }
                $ret = $self->remove_device_files($fileName);
                if ($ret) {
                    VCSAG_LOG_MSG("E", "Error removing the device file for group $self->{'name'} ", 67, "$self->{'name'}");
                    $self->_exit(0);
                }
            }

            $count = $self->get_count();
            VCSAG_LOG_MSG("N", "issuing establish/resume of group $self->{'name'}", 69, "$self->{'name'}");
            my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $count establish", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "Establish=link for the group $self->{'name'} failed; error code is $ret", 70, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }
        }


	if ($self->{'consistency'} eq "Enabled") {
		$count = $self->get_count();
		VCSAG_LOG_MSG("N", "re-enable consistency for group $self->{'name'}", 71, "$self->{'name'}");
		$ret = system("$symcg $self->{'type'} $self->{'name'} -noprompt -i 5 -c $count enable");
		if ($ret) {
			VCSAG_LOG_MSG("N", "re-enabling consistency on group $self->{'name'} failed. Administrative intervention required.", 72, "$self->{'name'}");
			$self->_exit(0);
		}
	}
        system("$symcfg sync");
        if ($ret) {
            $ret = $ret / 256;
            VCSAG_LOG_MSG("E", "Synching of database failed; error code is $ret", 81, "$ret");
            $self->_exit(0);
        }
        sleep(3);
        $self->get_state();
        if ($self->{'state'} eq "RW") {
            VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
            $self->vxvm_scan();
            $self->create_lockfile();
            if ($self->{'is_vmware'}) {
                $self->vmware_online();
            } else { 
                $self->cvm_import();
            }
            $self->reload_devicemap();
            $self->_exit(0);
        } else {
            VCSAG_LOG_MSG("E", "failover of group $self->{'name'} failed , devices are not read-write enabled", 80, "$self->{'name'}");
            $self->_exit(0);
        }
    } else {
        if ( (!$self->exists_swapfile()) || ($self->{'swaproles'} == 0)) {
# No Swap personality and normal failover.
            VCSAG_LOG_MSG("N", "issuing failover of group $self->{'name'}", 18, "$self->{'name'}");
            my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $failovercount failover 2>&1", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "failover of device group $self->{'name'} failed; error code is $ret", 42, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }
            sleep(3);
            $self->get_state();
            if ($self->{'state'} eq "RW") {
                VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
                $self->vxvm_scan();
                $self->create_lockfile();
                if ($self->{'is_vmware'}) {
                    $self->vmware_online();
                } else { 
                    $self->cvm_import();
                }
                $self->reload_devicemap();
                $self->_exit(0);
            } else {
                VCSAG_LOG_MSG("E", "failover of group $self->{'name'} failed, devices are not read-write enabled", 80, "$self->{'name'}");
                $self->_exit(0);
            }
        }

# For swap devices, if mode is sync.
        if ( !$self->get_mode ()) {
            VCSAG_LOG_MSG("N", "issuing failover/role swap of device group $self->{'name'}", 43, "$self->{'name'}");
            my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $failovercount failover -establish 2>&1", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "failover of device group $self->{'name'} failed; error code is $ret", 42, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }
            sleep(3);
            $self->get_state();
        } else {
            VCSAG_LOG_MSG("N", "issuing -force failover of device group $self->{'name'}", 44, "$self->{'name'}");
            my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -force -i 5 -c $failovercount failover 2>&1", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "failover of device group $self->{'name'} failed; error code is $ret", 42, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }

            $count = $self->get_count();
            $disabled_consistency = 0;
            if ($self->get_consistency()) {
                $disabled_consistency = 1;
                VCSAG_LOG_MSG("N", "disabling consistency for device group $self->{'name'}", 45, "$self->{'name'}");
                $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -i 5 -c $count -noprompt disable >$DEVNULL", 0, 1, $func);
                if ($ret) {
                    VCSAG_LOG_MSG("E", "Disabling the consistency for the device group $self->{'name'} failed; error code is $ret", 52, "$self->{'name'}", "$ret");
                    $self->_exit(0);
                }
            }
            sleep(3);
            $count = $self->get_count();
            VCSAG_LOG_MSG("N", "issuing role swap of device group $self->{'name'}", 46, "$self->{'name'}");
            $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $count swap >$DEVNULL", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "Role-swap of the device group $self->{'name'} failed; error code is $ret", 53, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }

            $count = $self->get_count();
            VCSAG_LOG_MSG("N", "issuing establish/resume of device group $self->{'name'}", 47, "$self->{'name'}");
            $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $count establish", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "Establish=link for the device group $self->{'name'} failed; error code is $ret", 54, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }

            $count = $self->get_count();
            if ($disabled_consistency) {
                VCSAG_LOG_MSG("N", "re-enable consistency for device group $self->{'name'}", 48, "$self->{'name'}");
                $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $count enable >$DEVNULL", 0, 1, $func);
                if ($ret) {
                    VCSAG_LOG_MSG("N", "re-enabling consistency on device group $self->{'name'} failed. Administrative intervention required.", 49, "$self->{'name'}");
                    $self->_exit(0);
                }
            }
            sleep(3);
            $self->get_state();
        }
    }
    if ($self->{'state'} eq "RW") {
        VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
        $self->vxvm_scan();
        $self->create_lockfile();
        if ($self->{'is_vmware'}) {
            $self->vmware_online();
        } else { 
            $self->cvm_import();
        }
        $self->reload_devicemap();
        $self->_exit(0);
    } else {
        VCSAG_LOG_MSG("E", "failover of group $self->{'name'} failed, devices are not read-write enabled", 80, "$self->{'name'}");
        $self->_exit(0);
    }
}

sub create_device_files() {
    my($self, $fileName, $LsymmID, $RsymmID, $RDFG) = @_;
    my ($i, $j, $k, @localDevices, @remoteDevices, $retVal) ;
    my ($keyLocal, $keyRemote);

    $retVal = 1;

    $keyLocal = "Local" . $RDFG . $LsymmID;
    $keyRemote = "Remote" . $RDFG . $RsymmID;

    foreach $k (keys (%{$self->{ARRAYDEVS}})) {
        foreach $j (keys (%{$self->{ARRAYDEVS}{$k}})) {
            if ($k eq $keyLocal) {
                push(@localDevices, $self->{ARRAYDEVS}{$k}{$j});
            } elsif ($k eq $keyRemote) {
                push(@remoteDevices, $self->{ARRAYDEVS}{$k}{$j});
            }
        }
    }
    
    if ($^O =~ /MSWin32/i) {
	    $$fileName = "$vcs_log" . "\\lock\\.deviceFile" . "_" . $self->{'name'} . "_" . $LsymmID . "_" . $RDFG . "_" . "$self->{'resname'}";
    } else {
	    $$fileName = "$vcs_log" . "/lock/.deviceFile" . "_" . $self->{'name'} . "_" . $LsymmID . "_" . $RDFG . "_" . "$self->{'resname'}";
    }
    open DEVICEFILE, ">$$fileName" || die ("Cannot open $$fileName file!\n");
    for ($i=0; $i<$#localDevices + 1; $i++) {
        print DEVICEFILE "$localDevices[$i] $remoteDevices[$i]\n";
    }
    close DEVICEFILE;
    $retVal = 0;

    return $retVal;
}

sub remove_device_files() {
    my($self, $fileName) = @_;
    my $retVal = 0;

    unlink $fileName;
    return $retVal;
}

sub vmware_online {
	# Make sure that /proc/vmware/config/LVM/EnableResignature and
	# /proc/vmware/config/LVM/DisallowSnapshotLun both are set to 0
	# rescanning will be done by parent VMFS resource.
    my $cmd = "esxcfg-advcfg -s 0 /LVM/EnableResignature";
    system ($cmd);
    $cmd = "esxcfg-advcfg -s 0 /LVM/DisallowSnapshotLun";
    system ($cmd);

    return 0;
}

sub get_lockfilename {
    my ($self) = shift;
    my $typ;
    my $lock_file;
    my $lock_dir = ""; 

    if (!defined($self->{'resource_type'}) || 
            (length($self->{'resource_type'}) == 0)) {

        $typ = `$ag_i18n_inc::HARES -value $self->{'resname'} Type`;
        chomp $typ;
    } else {
        $typ = $self->{'resource_type'};
    }

    if ($^O =~ /MSWin32/i) {
        $lock_dir = $ag_i18n_inc::VCSLOG . "\\lock\\" . $typ . "\\";
    } else {
        $lock_dir = $ag_i18n_inc::VCSLOG . "/lock/" . $typ . "/";
    }

    $lock_file = $lock_dir . ".VCS_" . $typ . "_" . $self->{'resname'} . 
        "_Online_Lockfile.DO-NOT-REMOVE";

    return $lock_file;
}


sub create_lockfile {
    my ($self) = shift;
    my ($lock_file) = $self->get_lockfilename();

    my (undef, $lock_dir, undef) = fileparse($lock_file);

    if (! -d $lock_dir) {
        mkdir  $lock_dir or 
            VCSAG_LOG_MSG("E", "Could not create lock directory $lock_dir",
                    1001, "$lock_dir");
    }

    open(LFH, ">$lock_file") or 
        VCSAG_LOG_MSG("E", "Could not create file: $lock_file", 1002,
                "$lock_file");

    close(LFH);
}


sub remove_lockfile {
    my($self) = shift;
    my ($lock_file) = $self->get_lockfilename();

    if (-f "$lock_file") {
        unlink("$lock_file") or 
            VCSAG_LOG_MSG("E", "Could not delete file: $lock_file", 1003,
                    "$lock_file");
    }
}


sub exists_lockfile {
    my ($self) = shift;
    my $retval = 0;
    my ($lock_file) = $self->get_lockfilename();

    if (-f "$lock_file") {
        $retval = 1;
    }
    return $retval;
}


sub get_swapfilename {
    my ($self) = shift;
    my $swap_file;
    my $swap_dir = ""; 
    my $typ;

    if (!defined($self->{'resource_type'}) || 
            (length($self->{'resource_type'}) == 0)) {

        $typ = `$ag_i18n_inc::HARES -value $self->{'resname'} Type`;
        chomp $typ;
    } else {
        $typ = $self->{'resource_type'};
    }

    if ($^O =~ /MSWin32/i) {
        $swap_dir = $ag_i18n_inc::VCSLOG . "\\lock\\" . $typ . "\\";
    } else {
        $swap_dir = $ag_i18n_inc::VCSLOG . "/lock/" . $typ . "/";
    }


    $swap_file = $swap_dir . ".VCS_" . $typ . "_" . $self->{'resname'} . 
        "_Dynamic_SwapOK.DO-NOT-REMOVE";

    return $swap_file;
}

sub create_swapfile {
    my ($self) = shift;
    my ($swap_file) = $self->get_swapfilename();

    my (undef, $swap_dir, undef) = fileparse($swap_file);

    if (! -d $swap_dir) {
        mkdir  $swap_dir or
            VCSAG_LOG_MSG("E", "Could not create directory $swap_dir", 1004,
                    "$swap_dir");;
    }

    open(LFH, ">$swap_file") or 
        VCSAG_LOG_MSG("E", "Could not create file: $swap_file", 1005, 
                "$swap_file");

    close(LFH);
}


sub remove_swapfile {
    my($self) = shift;
    my ($swap_file) = $self->get_swapfilename();

    if (-f "$swap_file") {
        unlink("$swap_file") or 
            VCSAG_LOG_MSG("E", "Could not delete file: $swap_file", 1006, 
                    "$swap_file");
    }
}


sub exists_swapfile {
    my ($self) = shift;
    my $retval = 0;
    my ($swap_file) = $self->get_swapfilename();

    if (-f "$swap_file") {
        $retval = 1;
    }
    return $retval;
}

#
# Return 1 if a parent resource of a parent service group is ONLINE
#

sub is_parent_online {
    my ($self) = shift;
    my ($p_res, $ret, $tmpstr, $sys);
    my @p_res_list = ();
    my $is_online = 0;

    $tmpstr = `$ag_i18n_inc::HASYS -nodeid`;
    chomp $tmpstr;
    $tmpstr =~ s/^.*Node name:[\s+](.*)[\s+]Node ID:.*$/$1/ ;
    $self->{'nodename'} = $sys = $tmpstr;

# Are there any resources that depend on this resource
    @p_res_list = $self->get_res_deps();

    if ($#p_res_list >= 0) { # Parent resources were found
        foreach $p_res (@p_res_list) {

            $ret = system("$ag_i18n_inc::HARES -wait $p_res Probed 1 "
                    ." -sys $sys -time 120 2>$DEVNULL");
            if ($ret == 0) {
                my ($st) = split('\n', `$ag_i18n_inc::HARES -state $p_res -sys $sys`);
                chomp $st;
                if ($st =~ /ONLINE/) {
                    $is_online = 1;
                    last;
                }
            }
        }
    } else { # No parent resources
        if ($self->is_parent_sg_res_online()) {
            $is_online = 1;
        }
    }

    return $is_online;
}

#
# Return 1 if at least one OnOff/OnOnly resource from a parent service
# group is online.
#

sub is_parent_sg_res_online {
    my ($self) = shift;
    my ($sg_name, $p_sg_name, $line, $ret);
    my @lines = ();
    my @fields = ();
    my @valid_dep_lines = ();
    my $is_online = 0;
    my $sys = $self->{'nodename'};

#
# Get parent service groups that have an online local/online global
# dependency to our service group
#

    $sg_name = `$ag_i18n_inc::HARES -value $self->{'resname'} Group`;
    chomp $sg_name;

    (@lines) = split('\n', `$ag_i18n_inc::HAGRP -dep $sg_name`);
    chomp @lines;

# VCS specific code --->
    (@valid_dep_lines) = grep (/online[\s+](local|global)/, @lines );

    foreach $line (@valid_dep_lines) {
        (@fields) = split('\s+', $line);

        if ($fields[1] eq $sg_name) {
            $p_sg_name = $fields[0];
        } else {
            next;
        }

		#
		# If at least one OnOff or OnOnly probed resource in the ParentSG
		# has gone online, then the SRDF resource can be online as well
		#

        $ret = system("$ag_i18n_inc::HAGRP -wait $p_sg_name Probed 1 -sys $sys"
                . " -time 120 2>$DEVNULL");

        if ($ret == 0) {
            my @p_on_res_list = $self->get_res_list($p_sg_name, "ONLINE");

			#
			# At least 1 online resource was found. Note: $#X == 0
			# implies @X has 1 element
			#
            if ($#p_on_res_list >= 0){

				#
				# We found at least one ONLINE resource. If it is of type 
				# OnOff or OnOnly, we can continue to remain ONLINE
				#   
                foreach my $p_res (@p_on_res_list) {
                    my $p_type = `$ag_i18n_inc::HARES -value $p_res Type`;
                    chomp $p_type;
                    my $p_oprn_type = `$ag_i18n_inc::HATYPE -value $p_type Operations`;
                    chomp $p_oprn_type;
                    if ($p_oprn_type =~ /None/i ) {
                        next;
                    }

                    $is_online = 1;
                    last;
                }
            }
        }
        last if ($is_online == 1);
    }

    return $is_online;
}

#
# Returns the names of resources that belong to a given group and
# are in particular state
#
sub get_res_list {
    my ($self) = shift;
    my ($sg_name, $res_state) = @_;
    my @retarr = ();
    my @tmparr = ();
    my ($ret, $nodename, $resname);
    my $func = "get_res_list";

    $ret = run_cmd("$ag_i18n_inc::HARES -list Group=$sg_name State=$res_state", 
            \@tmparr, 1, $func);

    if (!$ret) {
        foreach (@tmparr) {
            /^#/ && next;
            ($resname, $nodename)  = split;

            if ((!grep(/^${resname}$/, @retarr)) &&
                    ($nodename =~ /$self->{'nodename'}/)) {

                push (@retarr, $resname);
            }
        }
    }
    return @retarr;
}


#
# Returns the names of parent resources
#
sub get_res_deps {
    my ($self) = shift;

    my ($sg_name, $res_state) = @_;
    my @retarr = ();
    my @tmparr = ();
    my ($ret, $grpname, $p_resname, $c_resname);
    my $func = "get_res_deps";

    $ret = run_cmd("$ag_i18n_inc::HARES -dep $self->{'resname'}", \@tmparr, 1, $func);

    if (!$ret) {
        foreach (@tmparr) {
            /^#/ && next;
            ($grpname, $p_resname, $c_resname)  = split;

			#
			# Note: with VCSOne, the 'resname' would be <grpname>.<resname>
			# whereas the $c_resname would only have the trailing <resname> 
			# part
			#

            if ((!grep(/^${p_resname}$/, @retarr)) && 
                    ( $c_resname =~ /$self->{'resname'}$/ )) {
                push (@retarr, $p_resname);
            }
        }
    }

    return @retarr;
}


sub run_cmd {
    my ($cmd, $output_ref, $ignore_failure, $func) = @_;
    my $retval = 0;

    VCSAG_LOGDBG_MSG(2, "$func: Issuing command: [$cmd]");
    my $typ = ref($output_ref);
    if (!$typ) { # output is not a ref. So caller is ot interested in output.
        system($cmd);
        return $?;
    }

    my $cmd_txt = $cmd;
    if (1 == $ignore_failure) {
        $cmd_txt .= " 2>$DEVNULL";
    }
    if ($typ eq "ARRAY") {
        @{$output_ref} = `$cmd_txt`;
        chomp @{$output_ref};
    } else {
        $$output_ref = `$cmd_txt`;
        chomp ${$output_ref};
    }

    $retval = $?;
    if ($retval != 0) {
        unless (1 == $ignore_failure ) {
            $cmd_txt =~ s/\"//g;
            VCSAG_LOG_MSG("W", "Command: \'$cmd_txt\' failed !", 1007, "$cmd_txt");
        }
    }
    return $retval;
}

sub __hack_se7_0_X_bug1 {
    my ($self, $recv_ret, @out) = @_;
    my $pattern = "The device is already in the requested RDF state";
    my $ret;

    VCSAG_LOGDBG_MSG(1, "Entering __hack_se7_0_X_bug1 ");

    if ($recv_ret != 1) {
        $ret = $recv_ret;
        VCSAG_LOGDBG_MSG(1, "__hack_se7_0_X_bug1 :: VALID RET");
    } else {
        chomp @out;
        $ret = 18 if (!grep(/$pattern/, @out)) ;
        VCSAG_LOGDBG_MSG(1, "__hack_se7_0_X_bug1 :: 7.0.X BUG Found");
    }
    return $ret;
}


# Deletes swap file
sub init {
    my ($res) = @_;
    $res->remove_swapfile();
}

# returns '-cg' if group is a composite group, '-g' otherwise
sub get_cg_options {
    my ($self) = shift;
    return ($self->{'iscompositegroup'}) ? "-cg" : "-g";
}

sub get_symapi_servers
{
    my ($resname, $symhome) = @_;
    return if ($OS eq "Windows NT");
    my $func = "get_symapi_servers";

	#
	# Get the list of SYMAPI_SERVERs if populated.
	#
    VCSAG_LOGDBG_MSG(1, "get_symapi_servers(): reading value of SymapiServers");
    my $cmd_op;
    run_cmd("$ag_i18n_inc::HARES -value $resname SymapiServers", \$cmd_op, 1, $func);
    chomp $cmd_op;
    my @symapi_servers = split(/\s+/, $cmd_op);
    if (@symapi_servers) {
        my $symapi_server = "";
        foreach my $sym_serv (@symapi_servers) {
            VCSAG_LOGDBG_MSG(1, "Checking symapi server: [$sym_serv]");
            $ENV{SYMCLI_CONNECT} = $sym_serv;
            run_cmd("$symhome/bin/symcfg list 2>/dev/null", \$cmd_op, 1, $func);
            VCSAG_LOGDBG_MSG(2, "cmd returned: [$cmd_op]");
            my @local_arrays = grep(/Local/, split(/\n/, $cmd_op));
            if (!@local_arrays) {
                VCSAG_LOGDBG_MSG(1, "Invalid entry.");
                delete $ENV{SYMCLI_CONNECT};
                next;
            }
            VCSAG_LOGDBG_MSG(1, "Agent will temporarily use [$sym_serv] as the remote symapi server for all operations in this entry point.");
            $symapi_server = $sym_serv;
            last;
        }
        if (!$symapi_server) {
            VCSAG_LOG_MSG("E", "None of the Symapi Servers specified for this resource are currently available. Please check whether the correct Symapi Servers are configured in the the symapi configuration options file. The resource will not go online until then.", 1009);
            exit(0);
        }
    }
    else {
        VCSAG_LOGDBG_MSG(1, "No symapi server configured. Using local mode.");
    }
}

sub get_rdf_groups
{
    my ($self, $lines_ref) = @_;
    my @lines = @$lines_ref;

    VCSAG_LOGDBG_MSG(1, "get_rdf_groups: Entered.");
    my $rdfs = ();
    my $rdfinfo = {};
    my $old_rdfdata = 0;
    my $r2_rdfg = "";
    my $r2_rdfg_count = 0;
    my $preferred_site_match = 0;
    foreach my $line (@lines) {
        chomp $line;
        next if ($line =~ /^$/);
        if ($line =~ /Device Group RDF Information/) {
            if ($old_rdfdata) {
                $rdfs->{$rdfinfo->{'symid'}} = $rdfinfo;
                if ($rdfinfo->{'type'} eq "R2") { # Assuming that the local device will be R2 in only one rdfg.
                    $self->{'preferredsite'} = $rdfinfo->{'symid'};
                }
                $rdfinfo = {};
            }
            $old_rdfdata = 1;
            next;
        }
        if ($line =~ /RDF Type\s+:\s+(\S+)/) {
            $rdfinfo->{'type'} = $1;
            next;
        }
        if ($line =~ /RDF \(RA\) Group Number\s+:\s+(\S+)/) {
            $rdfinfo->{'rdfg'} = $1;
            next;
        }
        if ($line =~ /Remote Symmetrix ID\s+:\s+(\S*)/) {
            $rdfinfo->{'symid'} = $1;
            next;
        }
        if ($line =~ /RDF Mode\s+:\s+(\S*)/) {
            $rdfinfo->{'mode'} = $1;
            next;
        }
        if ($line =~ /RDF Pair Configuration\s+:\s+(\S*)/) {
            $rdfinfo->{'pairconfig'} = $1;
            next;
        }
        if ($line =~ /RDF STAR Mode\s+:\s+(\S*)/) {
            $rdfinfo->{'starmode'} = $1;
            next;
        }
        if ($line =~ /^\s+Device RDF State\s+: (.*)/) {
            $rdfinfo->{'rdfstate'} = $1;
            $rdfinfo->{'rdfstate'} =~ s/\s+/ /g;
            next;
        }
        if ($line =~ /Remote Device RDF State\s+: (.*)/) {
            $rdfinfo->{'remotestate'} = $1;
            $rdfinfo->{'remotestate'} =~ s/\s+/ /g;
            next;
        }
        if ($line =~ /RDF Pair State .*: (.*)/) {
            $rdfinfo->{'pairstate'} = $1;
            $rdfinfo->{'pairstate'} =~ s/\s+/ /g;
            next;
        }
    }
    if ($old_rdfdata) {
        $rdfs->{$rdfinfo->{'symid'}} = $rdfinfo;
        if ($rdfinfo->{'type'} eq "R2") { # Assuming that the local device will be R2 in only one rdfg.
            $self->{'preferredsite'} = $rdfinfo->{'symid'};
        }
        $rdfinfo = {};
    }
    if ($self->{'preferredsite'}) {
        VCSAG_LOGDBG_MSG(1, "get_rdf_groups: Found R2 RDFG.");
    } else {
        VCSAG_LOGDBG_MSG(1, "get_rdf_groups: No R2 RDFG.");
    }
    $self->{'rdfg'} = $rdfs;
    $self->{'preferredrdfg'} = $rdfs->{$self->{'preferredsite'}}->{'rdfg'};
    $self->{'rdfgopt'} = "-rdfg $self->{'preferredrdfg'}" if ($self->{'preferredrdfg'});
}

sub get_symms {

    my ($self, $lsymms_ref, $rsymms_ref) = @_;
    my @tmparr = ();
    my @lsid_lines;
    my @rsid_lines;
    my @lines;
    my $func = "get_symms";
	my $mode;
    $self->{'concurrent'} = 0;

    SRDFAgent::get_symapi_servers($self->{'resname'}, $self->{'home'});

    @{$lsymms_ref} = ();
    @{$rsymms_ref} = ();

    my $showcmd = ($self->{'iscompositegroup'} != 0) ? 
        "$symcg show $self->{'name'}" : 
        "$symdg show $self->{'name'}";
    if (SRDFAgent::run_cmd("$showcmd", \@tmparr, 1, $func)) {
        $self->_exit(0);
    }

    @lines = grep (/Symmetrix ID/, @tmparr);
    if ($OS ne "Windows NT") {
        VCSAG_LOGDBG_MSG(1, "get_symms Checking for concurrent...");
        my @rdf_groups = grep(/Device Group RDF Information/, @tmparr);
        my $num_rdfg = @rdf_groups;
        if ($num_rdfg > 1) {
            $self->{'concurrent'} = 1;
            VCSAG_LOGDBG_MSG(1, "Concurrent SRDF setup detected.");
            $self->get_rdf_groups(\@tmparr); # Create a map of Remote SymIds to RA Group No. Also sets the preferred site rdfg no.
        } else {
            $self->{'concurrent'} = 0;
        }
    }
# Find the Group Type for faster decision making.
    VCSAG_LOGDBG_MSG(1, "get_symms Checking for fast failover...");
    my $std_context = 0;
    $self->{'dev_state'} = "";
    foreach my $line (@tmparr) {
        $line =~ s/^\s+//;
        $line =~ s/\s+$//;
        if ($line =~ /Group Type\s+:\s+(\S+)/) {
            $self->{'grp_type'} = $1;
            next;
        }
        if ($line =~ /Standard \(STD\) Devices/) {
            $std_context = 1;
            next;
        }
        if ($std_context) {
            if ($line =~ /\}/) {
                $std_context = 0;
                next;
            }
            my @tokens = split(/\s+/, $line);
            if ($#tokens == 5 && $tokens[$#tokens] =~ /^\d+$/) {
                if ($self->{'dev_state'} eq "") {
                    $self->{'dev_state'} = $tokens[4];
                } else {
                    if ($self->{'dev_state'} ne $tokens[4]) {
                        $self->{'dev_state'} = "mix";
                    }
                }
            }
        }
		if ($line =~ /RDF Mode\s+:\s+(\S+)/) {
			$mode = $1;
			next
		}
		# In case of concurrent setup SRDF agent can come online only on the Synchronous link
		if ($line =~ /RDF Pair State .*:\s+(.*)/) {
			if ( $self->{'concurrent'} == 1 && $mode eq "Synchronous") {
				$self->{'pairstate'} = $1;
				$self->{'pairstate'} =~ s/\s+/ /g;
			} elsif ( $self->{'concurrent'} == 0) {
				$self->{'pairstate'} = $1;
				$self->{'pairstate'} =~ s/\s+/ /g;
			}
		}
    }
#
# for composite groups, @lines contains lines of the form:
#   1) Symmetrix ID                                   : 000187400527
#      Remote Symmetrix ID    : 000187900584
# for rdf device groups, the @lines is of the form:
#   Symmetrix ID                                   : 000187400527
#      Remote Symmetrix ID    : 000187900584
#

    @lsid_lines =  grep(!/Remote Symmetrix ID/, @lines);
    @rsid_lines =  grep(/Remote Symmetrix ID/, @lines);

    if (scalar(@lsid_lines) <= 0 or scalar(@rsid_lines) <= 0) {
        $ENV{"LANG"}="$USERLANG";
        VCSAG_LOG_MSG("E", "Could not find any Symmetrix ID associated with group $self->{'name'}\n", 1008, "$self->{'name'}");
        $self->_exit(0);
    }

    foreach my $line (@lsid_lines) {
        my (undef, $sid) = split (/:/, $line);
# get rid of all whitespaces
        $sid =~ s/\s+//g;
        push (@{$lsymms_ref}, $sid) if (!grep(/$sid/, @{$lsymms_ref}));
    }

    foreach my $line (@rsid_lines) {
        my (undef, $sid) = split (/:/, $line);
# get rid of all whitespaces
        $sid =~ s/\s+//g;
        push (@{$rsymms_ref}, $sid) if (!grep(/$sid/, @{$rsymms_ref}));
    }

    return;
}

#
# Returns 0 if any command failed
# Returns 1 otherwise
#
sub _get_dynamic_devs {
    my ($self, $dyndevs_href, @symms) = @_;
    %{$dyndevs_href} = ();
    my $func = "_get_dynamic_devs";

    my ($ret);
    my @tmparr = ();

    foreach my $localsymm (@symms) {
        my @dyndevs = ();
        if (SRDFAgent::run_cmd("$symdev -dynamic list -sid $localsymm", \@tmparr, 1, $func)) {
            return 0 ;
        }
        foreach my $line (@tmparr) {
            if ($line =~ /RDF/) {
                my (@fields) = split('\s+', $line);
                push(@dyndevs, $fields[0]);
            }
        }
        $dyndevs_href->{$localsymm} = [ @dyndevs ];
        if ($#dyndevs == -1) {
            return 0;
        }

    }
    return 1;
}

#
# Returns 0 if any command failed
# Returns 1 otherwise
#
sub _get_grp_devs {
    my ($self,$type, $grpdevs_href, @local_symms) = @_;
    %{$grpdevs_href} = ();
    my $func = "_get_grp_devs";

    my @fields = ();
    my @tmparr = ();
    my $cmd;

    foreach my $localsymm (@local_symms) {
        my @grpdevs = ();

        $cmd = "$symrdf $type $self->{'name'} query $self->{'rdfgopt'}";
#
# Composite groups can span multiple symmetrix arrays. We need to find
# our devices on each array
#
        if ($self->{'iscompositegroup'} != 0) {
            $cmd .= " -sid $localsymm";
        }

        if (SRDFAgent::run_cmd("$cmd", \@tmparr, 1, $func)) {
            return 0 ;
        }

# Consider only lines that contain  /[SAC]\./  eg: S. or A. etc
        foreach my $line (@tmparr) {
            if ($line =~ /Group Number/) {
                next;
            }
            if ($line =~ /^Remote Symmetrix ID/) {
                next;
            }
            if ($line =~ /^Symmetrix ID/ or $line =~ /^DG's Symmetrix ID/) {
                next;
            }
            if ($line =~ /Source/) {
                next;
            }
            if (!($line =~/[SAC]\./)) {
                next;
            }
            (@fields) = split('\s+', $line);
            push(@grpdevs, $fields[1]);
        }
        $grpdevs_href->{$localsymm} = [ @grpdevs ];

        if ($#grpdevs == -1) {
            return 0;
        }

    }

    return 1;
}

#
# Check if all our devices are dynamic devices
# Returns undef if any command failed, 
# 1 if true, 0 on false
#
sub is_dynamic_grp_devs {
    my ($self, $type, @local_symms) = @_;

    my $retval = 0;
    my $found = 0;
    my %dyndevs_hash = (); 
    my %grpdevs_hash = ();


    if(!($self->_get_grp_devs($type, \%grpdevs_hash, @local_symms)) ||
            !($self->_get_dynamic_devs(\%dyndevs_hash, @local_symms))) {

        return 0;
    }

    foreach my $localsymm (@local_symms) {
        my @grpdevs = @{ $grpdevs_hash{$localsymm} };
        my @dyndevs = @{ $dyndevs_hash{$localsymm} };

# Trivial case wherein either list is empty
        if (($#grpdevs == -1) || ($#dyndevs == -1)) {
            return 0;
        }

		#
		# If any device from the group does not exist in the 
		# dynamic devices list, return 0;
		#

        foreach my $grpdev (@grpdevs) {
            if (!grep(/^$grpdev$/, @dyndevs)) {
                return 0;
            }
        }
    }

    return 1;
}

#
# check that all connected arrays are at or above the supported microcode level
# returns 1 if all arrays are at/above supported microcode level, 0 otherwise 
#
sub is_mcode_version_OK {
    my ($self, $lsymms_ref) = @_;
    my $func = "is_mcode_version_OK";

    my @fields = ();
    my @tmparr = ();

    if (SRDFAgent::run_cmd("$symcfg list", \@tmparr, 1, $func)) {
        return undef;
    }

	#
	# Create a string of the form: 
	#   "<local1 sid>|<local2 sid>|<remote1 sid>|<remote2 sid>"
	#

    my $srchpattern = @{$lsymms_ref};
    my @lines = grep (/$srchpattern/, @tmparr);

	#
	# Each line is of the form:
	#  '   000187400527 Local   DMX1000S  5671   16384  37 3577'
	#
    foreach my $l (@lines) {
        (@fields) = split('\s+', $l);

        if ($fields[4] < SUPPORTED_MCODE_VERSION) {
            return 0; 
        }
    }

    return 1;
}

#
# Return 1 if Dynamic RDF config is enabled for the arrays; 0 otherwise
#
sub is_symmstate_dynamic {
    my ($self,@local_symms) = @_;

    my $retval = 0;
    my @tmparr = ();
    my $drc_state_cnt = 0;
    my $func = "is_symmstate_dynamic";

    foreach my $localsymm (@local_symms) {
        if (SRDFAgent::run_cmd("$symcfg -sid $localsymm list -v", \@tmparr, 1, $func)) {
            return 0;
        }
        if (grep (/\s+Dynamic RDF Configuration State\s+:\s+Enabled/, @tmparr)){
            $drc_state_cnt++ ;
        }
    }

    if ( $drc_state_cnt == scalar(@local_symms)){
        $retval = 1;
    }

    return $retval;
}

#
# CLI versions supported are 6.3 and above
# Returns 1 on success, 0 on error
#
sub is_CLI_version_OK {
    my ($self) = shift;
    my @tmparr = ();
    my $func = "is_CLI_version_OK";


# To workaround the issue in perl on Windows, run the command with some
# argument. backticks fail when the ShortPath contains special characters. 
# But strangely succeeds if we provide an argument.
    if (SRDFAgent::run_cmd("$symcli -v", \@tmparr, 1, $func)) {
        return 0;
    }

    if (grep(/V6\.[3-9]/, @tmparr) || grep (/V7\./, @tmparr)) {
        return 1;
    } else {
        $ENV{"LANG"}="$USERLANG";
        my $name = $self->{'name'};
        VCSAG_LOG_MSG("N", "Group $name is enabled for role swap on failover, but SYMCLI 6.3 or above is not installed; devices will be treated as normal devices", "$name");
        $ENV{"LANG"}="C";
        return 0;
    }
}

sub lock_res {
    my($self) = shift;
    my ($resname) = $self->{'resname'};

    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    my ($para) = split('\n', `$ag_i18n_inc::HAGRP -value $grp Parallel`);
    return 0 if ($para != 1);

    my ($lockholder) = split('\n', `$ag_i18n_inc::HARES -value $resname VCSResLock`);
    my (@locsysarr) = split('\s+', `LC_ALL=C; export LC_ALL; $ag_i18n_inc::HASYS -nodeid`);
    my ($thisnode) = $locsysarr[5];
    if ($lockholder eq "") {
        `$ag_i18n_inc::HARES -lock $resname > /dev/null 2>&1`;
        my $retval = $?;
        return 0 if ($retval == 0);
        ($lockholder) = split('\n', `$vcs_home/bin/hares -value $resname VCSResLock`);
    }
    return $lockholder if ($lockholder ne $thisnode);
    return 0;
}

sub unlock_res {
    my($self) = shift;

    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    my ($para) = split('\n', `$ag_i18n_inc::HAGRP -value $grp Parallel`);
    return 0 if ($para != 1);

    `$ag_i18n_inc::HARES -unlock $self->{'resname'} > /dev/null 2>&1`;
}

1;
